# Copyright 2013 Janne H. Korhonen, Pekka Parviainen
#
# This file is part of best-w-tree.
# 
# best-w-tree is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Foobar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with best-w-tree.  If not, see <http://www.gnu.org/licenses/>.

from w_tree_algs import *
import sys
import argparse


parser = argparse.ArgumentParser()
parser.add_argument("-n", help="use at most N variables from the input", type=int)
parser.add_argument("-z", help="interpret 0.0 as minus infinity", action="store_true")
parser.add_argument("-m", help="initialise all keys at the beginning to avoid resizing DP tables", action="store_true")
parser.add_argument("w", help="tree-width bound", type=int)
parser.add_argument("scorefile", help="local scores file")
args = parser.parse_args()

w = args.w
n = 0

f = read(args.scorefile, args.z)
if args.n is None:
	n = len(f)
else:
	n = args.n
	f = f[:n]
	f = map(lambda list: list[:(2**n)], f)

sys.stderr.write("doing zeta transform on f...\n")

f_hat = map(zeta_max,f)

sys.stderr.write("running base step for DP...\n")

sys.stderr.write("initialising DP algorithm...\n")
sys.stderr.flush()

N = 2**n
if args.m:
	sys.stderr.write("pre-computing keys for DP...\n")
	sys.stderr.flush()
	g = init_g_keys(n, N, w)
else:
	g = init_no_g_keys(n, N, w)


base_step(f, f_hat, w)

for i in range(1,n-w):
	sys.stderr.write("running DP for size " + str(w+i) + "...\n")
	update_g(f_hat, w, i)
	sys.stderr.flush()

sys.stderr.write("DP done\n")
sys.stderr.flush()


best = float("-inf")
best_tree_head = []
best_X = 0

for sigma in permutations(range(n), w+1):
	X = 0
	for i in sigma:
		X += (1 << i)
	current = g[(sigma, X, (2**n) - 1)]
	if best <= current:
		best = current
		best_tree_head = sigma
		best_X = X

print "best DAG has score", best

parentsets = [[]] * n
scores = [0] * n

print "tree-decomposition:"
get_best_tree(f,f_hat,best_tree_head,best_X, N - 1, "", parentsets, scores)

print "parents and local scores:"
for i in range(n):
	print i, "<-", " ".join(map(str,parentsets[i])), "( score:", scores[i], ")"



