# Copyright 2013 Janne H. Korhonen, Pekka Parviainen
#
# This file is part of best-w-tree.
# 
# best-w-tree is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Foobar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with best-w-tree.  If not, see <http://www.gnu.org/licenses/>.

from itertools import combinations, permutations
import sys
import math

MINUS_INF = float("-inf")

# globals so that we do not have to pass these around all the time
# probably should make a bit more sensible architecture here

cdef int n = 0
cdef int N = 0
g = {}
subsets = []

def read(scorefile, z):
	data = open(scorefile,'r')
	f = []
	for line in data:
		scores = line.split()
		for i in range(len(scores)):
			if (scores[i] == "0.0" and z) or scores[i] == "-":
				scores[i] = MINUS_INF
			else:
				scores[i] = float(scores[i])
		f.append(scores)
	return f


# takes a list f of length 2^n, returns the zeta transform of the list in max semi-ring
def zeta_max(f):
	N = len(f)
	n = int(math.log(N,2))
	y0 = f
	y1 = [0] * N
	for j in xrange(n):
		bitmask = 1 << j
		for a in xrange(N):
			if a & bitmask:
				y1[a] = max(y0[a], y0[a ^ bitmask])
			else:
				y1[a] = y0[a]
		y0, y1 = y1, [0] * N
	return y0


def init_no_g_keys(np,Np,w):
	global N
	global n
	global g
	global subsets
	
	N = Np
	n = np
	
	subsets = all_subsets()
	g = {}
	return g


def init_g_keys(np,Np,w):
	global N
	global n
	global g
	global subsets
	
	N = Np
	n = np
		
	subsets = all_subsets()
	
	keylist = [None] * ((2**(w+1)) * (2**(n-w-1)) * math.factorial(w+1) * binomial(n,w+1))
	
	list_ind = 0	
	
	for sigma in permutations(range(n),w+1):
		X = 0
		for i in sigma:
			X += (1 << i)
		for S in subsets[X]:
			for CM in subsets[X ^ (N-1)]:
				keylist[list_ind] = (sigma,S, CM ^ (N-1))
				list_ind += 1
				
	g = dict.fromkeys(keylist,0)
	return g
		

def base_step(f,f_hat,w):
	
	cdef double g_current
	
	
	cdef int X, S, Pb
	
	for X_list in combinations(range(n), w+1):
		X = 0
		for i in X_list:
			X += (1 << i)
		for S in subsets[X]:
			if S != 0: # S == 0 will not be used further on
				for sigma in permutations(X_list):
					Pb = 0
					g_current = 0
			
					for i in sigma:
						if S & (1 << i):
							g_current += f_hat[i][Pb]
						Pb += (1 << i)
			
					g[(sigma, S, X)] = g_current
	return n, N, g


def update_g(f_hat, int w, int extra_m):
		
	F = compute_Fs(f_hat, w, extra_m)
	
	supersets = all_supersets(w + 1, w + 1 + extra_m)
	
	partitions = all_partitions(extra_m)
	
	cdef int X, S, M
	cdef int Z, SmZ, Pb,L, R, M1, M2
	
	cdef double val_i, val_Z, val_K, val_J
	
	gl = g
	subsetsl = subsets
	
	
	if extra_m == 1:
		for sigma in permutations(range(n), w+1):
			X = 0
			for i in sigma:
				X += (1 << i)
			for S in subsetsl[X]:
				for M in supersets[X]:
					# g[(sigma, S, M)] = K(f_hat, sigma, S, M, F)
					val_K = float("-inf")
					for i in sigma:
						gamma = tuple(x for x in sigma if x != i)
						val_i = float("-inf")
						
						for Z in subsetsl[S & ~ (1 << i)]:
							
							val_Z = F[(gamma, Z, M & ~ (1 << i))]
					
							SmZ = S ^ Z
							Pb = 0
							for j in sigma:
								if SmZ & (1 << j):
									val_Z += f_hat[j][Pb]
								Pb += (1 << j)
							val_i = max(val_i, val_Z)
						val_K = max(val_K, val_i)			
					
					gl[(sigma, S, M)] = val_K
	else:
		for sigma in permutations(range(n), w+1):
			X = 0
			for i in sigma:
				X += (1 << i)
			for S in subsetsl[X]:
				for M in supersets[X]:
					# g[(sigma, S, M)] = max(J(sigma, S, M, extra_m), K(f_hat, sigma, S, M, F))
					val_K = float("-inf")
					for i in sigma:
						gamma = tuple(x for x in sigma if x != i)
						val_i = float("-inf")
						
						for Z in subsetsl[S & ~ (1 << i)]:
					
							val_Z = F[(gamma, Z, M & ~ (1 << i))]
							
							SmZ = S ^ Z
							Pb = 0
							for j in sigma:
								if SmZ & (1 << j):
									val_Z += f_hat[j][Pb]
								Pb += (1 << j)
							val_i = max(val_i, val_Z)
						val_K = max(val_K, val_i)
					
					
					val_J = float("-inf")
					L = M - X
					for R in partitions[L]:
						for Z in subsetsl[S]:
							val_J = max(val_J, gl[sigma, Z, X | R] + gl[sigma, S-Z, X | (L ^ R)])
					
					gl[(sigma, S, M)] = max(val_J,val_K)
	del F
	del supersets
	del partitions


# def J(sigma, int S, int M, int extra_m):
# 		
# 	cdef int X, Z, L, R, M1, M2
# 	
# 	X = 0
# 	for i in sigma:
# 		X += (1 << i)
# 	val = float("-inf")
# 	
# 	for Z in subsets[S]:
# 		L = M - X
# 		for R in subsets[L]:
# 			if R != 0 and R != L:
# 				
# 				M1 = X | R 
# 				M2 = X | (L ^ R)
# 				val = max(val, g[sigma, Z, M1] + g[sigma, S-Z, M2])
# 	
# 	return val
# 
# 
# def K(f_hat, sigma, int S, int M, F):
# 	
# 	cdef int Z, SmZ, Pb
# 	
# 	val = float("-inf")
# 	for i in sigma:
# 		gamma = tuple(x for x in sigma if x != i)
# 		val_i = float("-inf")
# 				
# 		for Z in subsets[S]:
# 		
# 			val_Z = F[(gamma, Z & ~ (1 << i), M & ~ (1 << i))]
# 			
# 			SmZ = S ^ Z
# 			Pb = 0
# 			for j in sigma:
# 				if SmZ & (1 << j):
# 					val_Z += f_hat[j][Pb]
# 				Pb += (1 << j)
# 			val_i = max(val_i, val_Z)
# 		val = max(val, val_i)			
# 	
# 	return val

def get_best_tree(f,f_hat,sigma, S, M, prefix, parentsets, scores):
	
	X = 0
	for i in sigma:
		X = X | (1 << i)
	
	if X == M:
		print prefix, (sigma, bin(S), bin(M)), "leaf node, node DP score", g[(sigma, S, M)]
		
		Pb = 0
		for i in sigma:
			if S & (1 << i):
				if parentsets[i] != []: # this really should not happen
					sys.stderr.write("error: node already had a parent set: " + str(i) + " " + str(parentsets[i])+"\n") 
				scores[i] = f_hat[i][Pb]
				for Ps in subsets[Pb]:
					if f[i][Ps] == scores[i]:
						parentsets[i] = elements(Ps)
			Pb += (1 << i)
		
		return
	
	val_K = float("-inf")
	best_K_sigma = []
	best_K_S = 0
	best_K_M = 0
	
	val_J = float("-inf")
	best_J_S = 0
	best_J_M = 0
	
	
	
	# compute the best swap-type predecessor for the current node
	
	for i in sigma:
		gamma = tuple(x for x in sigma if x != i)
		
		for Z in subsets[S & ~ (1 << i)]:
			
			# first we compute the best "swap target" for i and Z
			
			Y = X & ~ (1 << i)
			K = M & ~ (1 << i)
			
			
			val_F = float("-inf")
			best_F_sigma = []
			best_F_S = 0
			best_F_M = 0
			
			for k in elements(K - Y):
				for j in range(len(gamma)+1):
					nu = gamma[:j] + (k,) + gamma[j:]
					if val_F < g[(nu, Z | (1 << k), K)]:
						val_F = g[(nu, Z | (1 << k), K)]
						best_F_sigma = nu
						best_F_S = Z | (1 << k)
						best_F_M = K
			
			# then we compute the parents for S \ Z
			
			SmZ = S ^ Z
			Pb = 0
			for j in sigma:
				if SmZ & (1 << j):
					val_F += f_hat[j][Pb]
				Pb += (1 << j)
			
			if val_K < val_F:
				val_K = val_F
				best_K_sigma = best_F_sigma
				best_K_S = best_F_S
				best_K_M = best_F_M

		
	# compute the best join-type predecessor for the current node

	
	L = M & ~ X
	if size(L) > 1:
		for R in subsets[L]:
			if R != 0 and R != L:
				for Z in subsets[S]:
					if val_J < g[sigma, Z, X | R] + g[sigma, S & ~ Z, X | (L & ~ R)]:
						val_J = g[sigma, Z, X | R] + g[sigma, S & ~ Z, X | (L & ~ R)]
						best_J_S = Z
						best_J_M = X | R
		
	if val_J < val_K: # best solution is swap
		print prefix, (sigma, bin(S), bin(M)), "swap node, node DP score", g[(sigma, S, M)]
		
		SmZ = S & ~ best_K_S
		Pb = 0
		for i in sigma:
			if SmZ & (1 << i):
				if parentsets[i] != []:
					sys.stderr.write("error: node already had a parent set: " + str(i) + " " + str(parentsets[i])+"\n")
				scores[i] = f_hat[i][Pb]
				for Ps in subsets[Pb]:
					if f[i][Ps] == scores[i]:
						parentsets[i] = elements(Ps)
			Pb += (1 << i)
		
		get_best_tree(f,f_hat, best_K_sigma, best_K_S, best_K_M, prefix + "  ", parentsets, scores)
		
	else: # best solution is join
		print prefix, (sigma, bin(S), bin(M)), "join node, node DP score", g[(sigma, S, M)]
		get_best_tree(f,f_hat, sigma, best_J_S, best_J_M, prefix + "  ", parentsets, scores)
		get_best_tree(f,f_hat,sigma, S - best_J_S, X | (M - best_J_M), prefix + "  ", parentsets, scores)



def compute_Fs(f_hat, w, extra_m):
	F = {}
	
	cdef int X, Z, M
	
	supersets = all_supersets(w, w + extra_m)
	
	for gamma in permutations(range(n), w):
		X = 0
		for i in gamma:
			X += (1 << i)
		for Z in subsets[X]:
			for M in supersets[X]:
				F[(gamma, Z, M)] = compute_F(f_hat, gamma, X, Z, M)
	return F

def compute_F(f_hat, gamma, int X, int Z, int M):
	# print gamma, bin(X), bin(Z), bin(M)
	
	val = float("-inf")
	for i in elements(M - X):
		val_i = float("-inf")
		for j in range(len(gamma)+1):
			nu = gamma[:j] + (i,) + gamma[j:]
			val_i = max(val_i, g[(nu, Z | (1 << i), M)])
		val = max(val,val_i)
	return val

# ----------------


def elements(M):
	elems = []
	for i in range(n):
		if ( (1 << i) & M > 0):
			elems.append(i)
	return elems

def size(S):
	s = 0
	while S:
		s = s + (1 & S)
		S = S >> 1
	return s

def mask(S, M):
	for i in range(n):
		if ((1 << i) & M):
			if not (1 & S):
				M = M - (1 << i)
			S = S >> 1
	return M


def all_subsets():
	cdef int S
	
	subsets = [[]] * (2**n)
	for T in range(2**n):
		T_subsets = []
		t = size(T)
		for subset_mask in range(2**t):
			S = mask(subset_mask,T)
			T_subsets.append(S)
		subsets[T] = T_subsets
	return subsets



def all_supersets(w,k):
	
	cdef int X, M
	
	supersets = {}
	for X_list in combinations(range(n),w):
		X = 0
		for i in X_list:
			X += (1 << i)
		supersets[X] = []
	
	for M_list in combinations(range(n),k):
		M = 0
		for i in M_list:
			M += (1 << i)
		for X_list in combinations(M_list, w):
			X = 0
			for i in X_list:
				X += (1 << i)
			supersets[X].append(M)
	
	return supersets



def all_partitions(k):
	cdef int X
	
	partitions = {}
	for X_list in combinations(range(n),k):
		X = 0
		for i in X_list:
			X += (1 << i)
		subsetsX = subsets[X]
		partitions[X] = [S for S in subsetsX if (S != 0 and S != X and subsetsX.index(S) <= subsetsX.index( S ^ X))]
	
	return partitions


def binomial(n, k):
	if k < 0 or k > n:
		return 0
	if k > n - k: # take advantage of symmetry
		k = n - k
	c = 1
	for i in range(1,k+1):
		c = c * (n - (k - i))
		c = c // i
	return c